using CSV
using DataFrames
using Plots
using StatsPlots
using LaTeXStrings

if (length(ARGS) < 1)
    println("Please use with \"input_file.csv\"")
end

processes = [
    "QED Process: 'ke->ke'",
    "QED Process: 'ke->kke'",
    "QED Process: 'ke->kkke'",
    "QED Process: 'ke->kkkke'",
    "QED Process: 'ke->kkkkke'",
]

function proc_to_n(str::AbstractString)
    parts = split(str, "'")
    parts = split(parts[2], "->")
    k_count = count(c -> c == 'k', parts[2])
    return k_count
end

function beautify_title(str::AbstractString)
    parts = split(str, "'")

    preprefix = parts[1]
    infix = parts[2]
    sufsuffix = parts[3]

    parts = split(infix, "->")

    prefix = parts[1]
    suffix = parts[2]

    k_count = count(c -> c == 'k', suffix)
    B_count = count(c -> c == 'B', suffix)

    if k_count == 1 || B_count == 1
        new_suffix = suffix
    elseif k_count >= 1
        new_suffix = replace(suffix, r"k+" => "k^$k_count")
    elseif B_count >= 1
        new_suffix = replace(suffix, r"B+" => "B^$B_count")
    end

    return "QED Compton Scattering Process " * L"%$prefix \rightarrow %$new_suffix" * sufsuffix
end

input_file = ARGS[1]
df = CSV.read(input_file, DataFrame)
n_inputs = df[:, "n_inputs"][1]
gpus = df.gpu_devices[1]
cpus = df.cpu_threads[1]

power = Int(round(log2(n_inputs)))

chunk_sizes = [
    "\$2^{10}\$",
    "\$2^{11}\$",
    "\$2^{12}\$",
    "\$2^{13}\$",
    "\$2^{14}\$",
    "\$2^{15}\$",
    "\$2^{16}\$",
    "\$2^{17}\$",
    "\$2^{18}\$",
    "\$2^{19}\$",
    "\$2^{20}\$",
]

best_times = Vector{Float64}()
best_times_std = Vector{Float64}()

# plotting with threads as x axis
for process_name in processes
    df_filt = filter(:process_name => x -> x == process_name, df)

    df_filt.cpu_ratio = df_filt.cpu_chunks ./ (df_filt.cpu_chunks .+ df_filt.gpu_chunks) .* 100.0
    df_filt.gpu_ratio = df_filt.gpu_chunks ./ (df_filt.cpu_chunks .+ df_filt.gpu_chunks) .* 100.0

    push!(best_times, minimum(df_filt.time))

    bar(chunk_sizes, df_filt.cpu_ratio, label = "workload completed by \$$(cpus)\$ CPU threads (%)")
    bar!(
        chunk_sizes,
        [100 for _ in chunk_sizes],
        label = "workload completed by $(gpus) GPUs (%)",
        fillto = df_filt.cpu_ratio,
    )

    plot!(
        #title = "$(beautify_title(process_name))\nComputing $(n_inputs) (\$2^{$(power)}\$) Matrix Elements",
        yscale = :linear,
        #xticks = [1024 4096 16384 65536 262144 1048576],
        ylim = (0, 105),
        legend = :outerbottom,
        legendcolumns = 1,
        legend_font_pointsize = 10,
        #size = (800, 600),
        ylabel = "contribution (%)",
        xlabel = "chunk size (#)",
    )

    savefig("full_node_chunk_size_$(proc_to_n(process_name))k_ratio.pdf")


    scatter(
        chunk_sizes,
        df_filt.rate,
        label = "total execution rate (\$s^{-1}\$)",
        #title = "$(beautify_title(process_name))\nComputing $(n_inputs) (\$2^{$(power)}\$) Matrix Elements",
        ylim = (0, :auto),
        #yscale = :log10,
        #xticks = [1024 4096 16384 65536 262144 1048576],
        legend = :outerbottom,
        legendcolumns = 1,
        legend_font_pointsize = 10,
        #size = (800, 600),
        ylabel = "rate (\$s^{-1}\$)",
        xlabel = "chunk size (#)",
        markersize = 7,
    )

    savefig("full_node_chunk_size_$(proc_to_n(process_name))k_rate.pdf")


    scatter(
        chunk_sizes,
        df_filt.time,
        yerror = df_filt.std,
        label = "total execution time (s)",
        #title = "$(beautify_title(process_name))\nComputing $(n_inputs) (\$2^{$(power)}\$) Matrix Elements",
        #xticks = [1024 4096 16384 65536 262144 1048576],
        ylim = (0, maximum(df_filt.time) * 1.05),
        legend = :outerbottom,
        legendcolumns = 1,
        legend_font_pointsize = 10,
        #size = (800, 600),
        ylabel = "time (s)",
        xlabel = "chunk size (#)",
        markersize = 7,
    )

    savefig("full_node_chunk_size_$(proc_to_n(process_name))k_time.pdf")

end


# plotting with process size as x axis
A100_rates = [2.530045276927587e9, 1.16972304616864e9, 2.0002725972692013e8, 3.495722925446318e7, 4.792187095617111e6]
CPU_32threads_rates =
    [3.2691139045711152e7, 1.1578342663759507e7, 3.1670680975577887e6, 731037.7069429948, 115001.5594731802]

theory_rates = (A100_rates .+ CPU_32threads_rates) .* 4

scatter(
    proc_to_n.(processes),
    best_times,
    label = "full node best achieved time (s)",
    #title = "QED N-Photon Compton Scattering\nComputing $(n_inputs) (\$2^{$(power)}\$) Matrix Elements",
    ylim = (0, maximum(best_times) * 1.05),
    legend = :outerbottom,
    legendcolumns = 1,
    legend_font_pointsize = 10,
    #size = (800, 600),
    ylabel = "time (s)",
    xlabel = "process size (#)",
    markersize = 7,
)

savefig("full_node_process_best_time.pdf")


scatter(
    proc_to_n.(processes),
    (n_inputs ./ best_times),
    label = "full node best achieved rate (\$s^{-1}\$)",
    #title = "QED N-Photon Compton Scattering\nComputing $(n_inputs) (\$2^{$(power)}\$) Matrix Elements",
    ylim = (0, maximum(n_inputs ./ best_times) * 1.05),
    legend = :outerbottom,
    legendcolumns = 1,
    legend_font_pointsize = 10,
    #size = (800, 600),
    ylabel = "rate (\$s^{-1}\$)",
    xlabel = "process size (#)",
    markersize = 7,
)

savefig("full_node_process_best_rate.pdf")



scatter(
    proc_to_n.(processes),
    [(n_inputs ./ best_times) theory_rates],
    label = ["full node best achieved rate (\$s^{-1}\$)" "theoretical rate from previous benchmarks (\$s^{-1}\$)"],
    #title = "QED N-Photon Compton Scattering\nComputing $(n_inputs) (\$2^{$(power)}\$) Matrix Elements",
    #ylim = (0, max(maximum(n_inputs ./ best_times), maximum(theory_rates)) * 1.05),
    yscale = :log10,
    legend = :outerbottom,
    legendcolumns = 1,
    legend_font_pointsize = 10,
    #size = (800, 600),
    ylabel = "rate (\$s^{-1}\$)",
    xlabel = "process size (#)",
    markersize = 7,
)

savefig("full_node_process_best_rate_plus_theory.pdf")
