using MetagraphOptimization
using CUDA
using UUIDs
using BenchmarkTools
using DataFrames
using CSV
using Dates

results_filename = "bench_results_reduction_steps_gpu.csv"

df = DataFrame(
    threads = Int[],
    process = String[],
    operations = Int[],
    cumulative_optimization_time = Float64[],
    graph_nodes = Int[],
    graph_edges = Int[],
    graph_ce = Float64[],
    graph_dt = Float64[],
    graph_ci = Float64[],
    cpu_st_t = Float64[],
    cpu_st_s = Float64[],
    cpu_mt_t = Float64[],
    cpu_mt_s = Float64[],
    cpu_mem = Float64[],
    gpu_t = Float64[],
    gpu_s = Float64[],
    gpu_mem = Float64[],
)

# if they exist, read existing results and append new ones
if isfile(results_filename)
    df = CSV.read(results_filename, DataFrame)
end

function log(x...)
    println(now(), " ", join(x, " ")...)
    return flush(stdout)
end

function bench(func, kernel!, inputs)
    # gpu part
    n = length(inputs)
    cu_inputs = CuVector(inputs)
    cu_outputs = CuVector{ComplexF64}()
    resize!(cu_outputs, n)
    ts = 32
    bs = Int(n / ts)
    bench = @benchmark begin
        @cuda threads = $ts blocks = $bs always_inline = true $kernel!($cu_inputs, $cu_outputs, $n)
        CUDA.device_synchronize()
    end gcsample = true samples = 20 evals = 1

    gpu_time = median(bench.times) / 1e9
    gpu_std = std(bench.times) / 1e9
    gpu_mem = bench.memory

    # cpu part
    single_thread = @benchmark $func.($inputs)
    multi_threaded = @benchmark Threads.@threads for i in eachindex($inputs)
        $func($inputs[i])
    end

    cpu_st_time = median(single_thread.times) / 1e9
    cpu_st_std = std(single_thread.times) / 1e9
    cpu_mt_time = median(multi_threaded.times) / 1e9
    cpu_mt_std = std(multi_threaded.times) / 1e9
    cpu_mem = std(single_thread.times)


    return (
        cpu_single_thread_time = cpu_st_time,
        cpu_single_thread_std = cpu_st_std,
        cpu_multi_thread_time = cpu_mt_time,
        cpu_multi_thread_std = cpu_mt_std,
        cpu_mem = cpu_mem,
        gpu_time = gpu_time,
        gpu_std = gpu_std,
        gpu_mem = gpu_mem,
    )
end

log("Available CUDA devices:")
for dev in CUDA.devices()
    display(dev)
end

# preparation of machine
machine = Machine(
    [
        MetagraphOptimization.NumaNode(
            0,
            1,
            MetagraphOptimization.default_strategy(MetagraphOptimization.NumaNode),
            -1.0,
            UUIDs.uuid1(),
        ),
    ],
    [-1.0;;],
)


# bench and produce data
n_inputs = 2^16
optimizer = ReductionOptimizer()
processes = [("ke->ke", 1), ("ke->kke", 1), ("ke->kkke", 1), ("ke->kkkke", 5)]

for (process_str, STEPSIZE) in processes
    n = 0
    opt_time_cum = 0

    process = parse_process(process_str, QEDModel())
    graph = gen_graph(process)
    inputs = Vector([gen_process_input(process) for _ in 1:n_inputs])

    get_compute_function(graph, process, machine)

    while true
        func = get_compute_function(graph, process, machine)
        kernel! = get_cuda_kernel(graph, process, machine)
        res = bench(func, kernel!, inputs)

        graph_properties = get_properties(graph)
        push!(
            df,
            (
                Threads.nthreads(),
                process_str,
                n,
                opt_time_cum,
                graph_properties.noNodes,
                graph_properties.noEdges,
                graph_properties.computeEffort,
                graph_properties.data,
                graph_properties.computeIntensity,
                res.cpu_single_thread_time,
                res.cpu_single_thread_std,
                res.cpu_multi_thread_time,
                res.cpu_multi_thread_std,
                res.cpu_mem,
                res.gpu_time,
                res.gpu_std,
                res.gpu_mem,
            ),
        )
        CSV.write(results_filename, df)

        if fixpoint_reached(optimizer, graph)
            break
        end

        opt_time_cum += @elapsed optimize!(optimizer, graph, STEPSIZE)
        n += STEPSIZE
    end
end

CSV.write(results_filename, df)

for (process_str, STEPSIZE) in [("AB->AB", 1), ("AB->ABBB", 1), ("AB->ABBBBB", 1)]
    n = 0
    opt_time_cum = 0

    process = parse_process(process_str, ABCModel())
    graph = parse_dag("input/$process_str.txt", ABCModel())
    inputs = Vector([gen_process_input(process) for _ in 1:n_inputs])

    get_compute_function(graph, process, machine)

    while true
        func = get_compute_function(graph, process, machine)
        kernel! = get_cuda_kernel(graph, process, machine)
        res = bench(func, kernel!, inputs)

        graph_properties = get_properties(graph)
        push!(
            df,
            (
                Threads.nthreads(),
                process_str,
                n,
                opt_time_cum,
                graph_properties.noNodes,
                graph_properties.noEdges,
                graph_properties.computeEffort,
                graph_properties.data,
                graph_properties.computeIntensity,
                res.cpu_single_thread_time,
                res.cpu_single_thread_std,
                res.cpu_multi_thread_time,
                res.cpu_multi_thread_std,
                res.cpu_mem,
                res.gpu_time,
                res.gpu_std,
                res.gpu_mem,
            ),
        )
        CSV.write(results_filename, df)

        if fixpoint_reached(optimizer, graph)
            break
        end

        opt_time_cum += @elapsed optimize!(optimizer, graph, STEPSIZE)
        n += STEPSIZE
    end
end

CSV.write(results_filename, df)
