"""
    get_compute_function(graph::DAG, instance, machine::Machine)

Return a function of signature `compute_<id>(input::input_type(instance))`, which will return the result of the DAG computation on the given input.
"""
function get_compute_function(graph::DAG, instance, machine::Machine)
    tape = gen_tape(graph, instance, machine)

    initCaches = Expr(:block, tape.initCachesCode...)
    assignInputs = Expr(:block, tape.inputAssignCode...)
    code = Expr(:block, expr_from_fc.(tape.computeCode)...)

    functionId = to_var_name(UUIDs.uuid1(rng[1]))
    resSym = eval(gen_access_expr(entry_device(tape.machine), tape.outputSymbol))
    expr = Meta.parse(
        "function compute_$(functionId)(data_input::$(input_type(instance))) $(initCaches); $(assignInputs); $code; return $resSym; end",
    )

    return expr
end

"""
    get_cuda_kernel(graph::DAG, instance, machine::Machine)

Return a function of signature `compute_<id>(input::CuVector, output::CuVector, n::Int64)`, which will return the result of the DAG computation of the input on the given output variable.
"""
function get_cuda_kernel(graph::DAG, instance, machine::Machine)
    tape = gen_tape(graph, instance, machine)

    initCaches = Expr(:block, tape.initCachesCode...)
    assignInputs = Expr(:block, tape.inputAssignCode...)
    code = Expr(:block, expr_from_fc.(tape.computeCode)...)

    functionId = to_var_name(UUIDs.uuid1(rng[1]))
    resSym = eval(gen_access_expr(entry_device(tape.machine), tape.outputSymbol))
    expr = Meta.parse(
        "function compute_$(functionId)(input_vector, output_vector, n::Int64)
            id = (blockIdx().x - 1) * blockDim().x + threadIdx().x
            if (id > n)
                return
            end
            @inline data_input = input_vector[id]
            $(initCaches)
            $(assignInputs)
            $code
            @inline output_vector[id] = $resSym
            return nothing
        end"
    )

    return expr
end

"""
    execute(graph::DAG, instance, machine::Machine, input)

Execute the code of the given `graph` on the given input values.

This is essentially shorthand for
```julia
tape = gen_tape(graph, instance, machine)
return execute_tape(tape, input)
```

See also: [`parse_dag`](@ref), [`parse_process`](@ref), [`gen_process_input`](@ref)
"""
function execute(graph::DAG, instance, machine::Machine, input)
    tape = gen_tape(graph, instance, machine)
    return execute_tape(tape, input)
end
