using QEDprocesses
using StaticArrays

const e = sqrt(4π / 137)

"""
    QEDModel <: AbstractPhysicsModel

Singleton definition for identification of the QED-Model.
"""
struct QEDModel <: AbstractPhysicsModel end

QEDbase.is_incoming(::Type{<:ParticleStateful{Incoming}}) = true
QEDbase.is_outgoing(::Type{<:ParticleStateful{Outgoing}}) = true
QEDbase.is_incoming(::Type{<:ParticleStateful{Outgoing}}) = false
QEDbase.is_outgoing(::Type{<:ParticleStateful{Incoming}}) = false

QEDbase.particle_direction(::Type{<:ParticleStateful{DIR}}) where {DIR <: ParticleDirection} = DIR()
QEDbase.particle_species(
    ::Type{<:ParticleStateful{DIR, SPECIES}},
) where {DIR <: ParticleDirection, SPECIES <: AbstractParticleType} = SPECIES()

_assert_particle_type_tuple(::Tuple{}) = nothing
_assert_particle_type_tuple(t::Tuple{AbstractParticleType, Vararg}) = _assert_particle_type_tuple(t[2:end])
_assert_particle_type_tuple(t::Any) =
    throw(InvalidInputError("invalid input, provide a tuple of AbstractParticleTypes to construct a GenericQEDProcess"))

struct GenericQEDProcess{INT, OUTT, INSP, OUTSP} <:
       AbstractProcessDefinition where {INT <: Tuple, OUTT <: Tuple, INSP <: Tuple, OUTSP <: Tuple}
    incoming_particles::INT
    outgoing_particles::OUTT

    incoming_spins_pols::INSP
    outgoing_spins_pols::OUTSP

    function GenericQEDProcess(
        in_particles::INT,
        out_particles::OUTT,
        in_sp::INSP,
        out_sp::OUTSP,
    ) where {INT <: Tuple, OUTT <: Tuple, INSP <: Tuple, OUTSP <: Tuple}
        _assert_particle_type_tuple(in_particles)
        _assert_particle_type_tuple(out_particles)

        # TODO: check in_sp/out_sp fits to particles (spin->fermion, pol->photon)

        return new{INT, OUTT, INSP, OUTSP}(in_particles, out_particles, in_sp, out_sp)
    end

    """
        GenericQEDProcess{INSP, OUTSP}(in_ph::Int, out_ph::Int, in_el::Int, out_el::Int, in_po::Int, out_po::Int, in_sp::INSP, out_sp::OUTSP)

    Convenience constructor from numbers of input/output photons, electrons and positrons.

    `in_sp` and `out_sp` are tuples of the spin and polarization configurations of the given particles. Their order is Photons, then Electrons, then Positrons for both incoming and outgoing particles.

    # TODO: make default for in_sp and out_sp available for convenience
    """
    function GenericQEDProcess(
        in_ph::Int,
        out_ph::Int,
        in_el::Int,
        out_el::Int,
        in_po::Int,
        out_po::Int,
        in_sp::INSP,
        out_sp::OUTSP,
    ) where {INSP <: Tuple, OUTSP <: Tuple}
        in_p = ntuple(i -> i <= in_ph ? Photon() : i <= in_ph + in_el ? Electron() : Positron(), in_ph + in_el + in_po)
        out_p = ntuple(
            i -> i <= out_ph ? Photon() : i <= out_ph + out_el ? Electron() : Positron(),
            out_ph + out_el + out_po,
        )
        return GenericQEDProcess(in_p, out_p, in_sp, out_sp)
    end
end

function spin_or_pol(
    process::GenericQEDProcess,
    type::Type{ParticleStateful{DIR, SPECIES, EL}},
    n::Int,
) where {DIR <: ParticleDirection, SPECIES <: AbstractParticleType, EL <: AbstractFourMomentum}
    i = 0
    c = n
    for p in particles(process, DIR())
        i += 1
        if p == SPECIES()
            c -= 1
        end
        if c == 0
            break
        end
    end

    if c != 0 || n <= 0
        throw(InvalidInputError("could not get $n-th spin/pol of $(DIR()) $species, does not exist"))
    end

    if DIR <: Incoming
        return process.incoming_spins_pols[i]
    elseif DIR <: Outgoing
        return process.outgoing_spins_pols[i]
    else
        throw(InvalidInputError("unknown direction $(DIR()) given"))
    end
end

QEDprocesses.incoming_particles(proc::GenericQEDProcess) = proc.incoming_particles
QEDprocesses.outgoing_particles(proc::GenericQEDProcess) = proc.outgoing_particles

function isphysical(proc::GenericQEDProcess)
    return (
        number_particles(proc, Incoming(), Electron()) + number_particles(proc, Outgoing(), Positron()) ==
        number_particles(proc, Incoming(), Positron()) + number_particles(proc, Outgoing(), Electron())
    ) && number_particles(proc, Incoming()) + number_particles(proc, Outgoing()) >= 2
end

function input_type(p::GenericQEDProcess)
    in_t = QEDcore._assemble_tuple_type(incoming_particles(p), Incoming(), SFourMomentum)
    out_t = QEDcore._assemble_tuple_type(outgoing_particles(p), Outgoing(), SFourMomentum)
    return PhaseSpacePoint{
        typeof(p),
        PerturbativeQED,
        PhasespaceDefinition{SphericalCoordinateSystem, ElectronRestFrame},
        Tuple{in_t...},
        Tuple{out_t...},
    }
end

ValueType = Union{BiSpinor, AdjointBiSpinor, DiracMatrix, SLorentzVector{Float64}, ComplexF64}

"""
    interaction_result(t1::Type{T1}, t2::Type{T2}) where {T1 <: QEDParticle, T2 <: QEDParticle}

For two given particle types that can interact, return the third.
"""
function interaction_result(t1::Type{T1}, t2::Type{T2}) where {T1 <: ParticleStateful, T2 <: ParticleStateful}
    @assert false "Invalid interaction between particles of types $t1 and $t2"
end

interaction_result(
    ::Type{ParticleStateful{Incoming, Electron, EL}},
    ::Type{ParticleStateful{Outgoing, Electron, EL}},
) where {EL <: AbstractFourMomentum} = ParticleStateful{Incoming, Photon, EL}
interaction_result(
    ::Type{ParticleStateful{Incoming, Electron, EL}},
    ::Type{ParticleStateful{Incoming, Positron, EL}},
) where {EL <: AbstractFourMomentum} = ParticleStateful{Incoming, Photon, EL}
interaction_result(
    ::Type{ParticleStateful{Incoming, Electron, EL}},
    ::Type{ParticleStateful{DIR, Photon, EL}},
) where {EL <: AbstractFourMomentum, DIR <: ParticleDirection} = ParticleStateful{Outgoing, Electron, EL}

interaction_result(
    ::Type{ParticleStateful{Outgoing, Electron, EL}},
    ::Type{ParticleStateful{Incoming, Electron, EL}},
) where {EL <: AbstractFourMomentum} = ParticleStateful{Incoming, Photon, EL}
interaction_result(
    ::Type{ParticleStateful{Outgoing, Electron, EL}},
    ::Type{ParticleStateful{Outgoing, Positron, EL}},
) where {EL <: AbstractFourMomentum} = ParticleStateful{Incoming, Photon, EL}
interaction_result(
    ::Type{ParticleStateful{Outgoing, Electron, EL}},
    ::Type{ParticleStateful{DIR, Photon, EL}},
) where {EL <: AbstractFourMomentum, DIR <: ParticleDirection} = ParticleStateful{Incoming, Electron, EL}

# antifermion mirror
interaction_result(
    ::Type{ParticleStateful{Incoming, Positron, EL}},
    t2::Type{<:ParticleStateful},
) where {EL <: AbstractFourMomentum} = interaction_result(ParticleStateful{Outgoing, Electron, EL}, t2)
interaction_result(
    ::Type{ParticleStateful{Outgoing, Positron, EL}},
    t2::Type{<:ParticleStateful},
) where {EL <: AbstractFourMomentum} = interaction_result(ParticleStateful{Incoming, Electron, EL}, t2)

# photon commutativity
interaction_result(
    t1::Type{ParticleStateful{DIR, Photon, EL}},
    t2::Type{<:ParticleStateful},
) where {EL <: AbstractFourMomentum, DIR <: ParticleDirection} = interaction_result(t2, t1)

# but prevent stack overflow
function interaction_result(
    t1::Type{ParticleStateful{DIR1, Photon, EL}},
    t2::Type{ParticleStateful{DIR2, Photon, EL}},
) where {DIR1 <: ParticleDirection, DIR2 <: ParticleDirection, EL <: AbstractFourMomentum}
    @assert false "Invalid interaction between particles of types $t1 and $t2"
end

"""
    propagation_result(t1::Type{T}) where {T <: QEDParticle}

Return the type of the inverted direction. E.g.
"""
propagation_result(::Type{ParticleStateful{Incoming, Electron, EL}}) where {EL <: AbstractFourMomentum} =
    ParticleStateful{Outgoing, Electron, EL}
propagation_result(::Type{ParticleStateful{Outgoing, Electron, EL}}) where {EL <: AbstractFourMomentum} =
    ParticleStateful{Incoming, Electron, EL}
propagation_result(::Type{ParticleStateful{Incoming, Positron, EL}}) where {EL <: AbstractFourMomentum} =
    ParticleStateful{Outgoing, Positron, EL}
propagation_result(::Type{ParticleStateful{Outgoing, Positron, EL}}) where {EL <: AbstractFourMomentum} =
    ParticleStateful{Incoming, Positron, EL}
propagation_result(::Type{ParticleStateful{Incoming, Photon, EL}}) where {EL <: AbstractFourMomentum} =
    ParticleStateful{Outgoing, Photon, EL}
propagation_result(::Type{ParticleStateful{Outgoing, Photon, EL}}) where {EL <: AbstractFourMomentum} =
    ParticleStateful{Incoming, Photon, EL}

"""
    types(::QEDModel)

Return a Vector of the possible types of particle in the [`QEDModel`](@ref).
"""
function types(::QEDModel)
    return [
        ParticleStateful{Incoming, Photon, SFourMomentum},
        ParticleStateful{Outgoing, Photon, SFourMomentum},
        ParticleStateful{Incoming, Electron, SFourMomentum},
        ParticleStateful{Outgoing, Electron, SFourMomentum},
        ParticleStateful{Incoming, Positron, SFourMomentum},
        ParticleStateful{Outgoing, Positron, SFourMomentum},
    ]
end

# type piracy?
String(::Type{Incoming}) = "Incoming"
String(::Type{Outgoing}) = "Outgoing"

String(::Type{PolX}) = "polx"
String(::Type{PolY}) = "poly"

String(::Type{SpinUp}) = "spinup"
String(::Type{SpinDown}) = "spindown"

String(::Incoming) = "i"
String(::Outgoing) = "o"

function String(::Type{<:ParticleStateful{DIR, Photon}}) where {DIR <: ParticleDirection}
    return "k"
end
function String(::Type{<:ParticleStateful{DIR, Electron}}) where {DIR <: ParticleDirection}
    return "e"
end
function String(::Type{<:ParticleStateful{DIR, Positron}}) where {DIR <: ParticleDirection}
    return "p"
end

"""
    caninteract(T1::Type{<:ParticleStateful}, T2::Type{<:ParticleStateful})

For two given `ParticleStateful` types, return whether they can interact at a vertex. This is equivalent to `!issame(T1, T2)`.

See also: [`issame`](@ref) and [`interaction_result`](@ref)
"""
function caninteract(
    T1::Type{<:ParticleStateful{D1, S1}},
    T2::Type{<:ParticleStateful{D2, S2}},
) where {D1 <: ParticleDirection, S1 <: AbstractParticleType, D2 <: ParticleDirection, S2 <: AbstractParticleType}
    if (T1 == T2)
        return false
    end
    if (S1 == Photon && S2 == Photon)
        return false
    end

    for (P1, P2) in [(T1, T2), (T2, T1)]
        if (P1 <: ParticleStateful{Incoming, Electron} && P2 <: ParticleStateful{Outgoing, Positron})
            return false
        end
        if (P1 <: ParticleStateful{Outgoing, Electron} && P2 <: ParticleStateful{Incoming, Positron})
            return false
        end
    end

    return true
end

function type_index_from_name(::QEDModel, name::String)
    if startswith(name, "ki")
        return (ParticleStateful{Incoming, Photon, SFourMomentum}, parse(Int, name[3:end]))
    elseif startswith(name, "ko")
        return (ParticleStateful{Outgoing, Photon, SFourMomentum}, parse(Int, name[3:end]))
    elseif startswith(name, "ei")
        return (ParticleStateful{Incoming, Electron, SFourMomentum}, parse(Int, name[3:end]))
    elseif startswith(name, "eo")
        return (ParticleStateful{Outgoing, Electron, SFourMomentum}, parse(Int, name[3:end]))
    elseif startswith(name, "pi")
        return (ParticleStateful{Incoming, Positron, SFourMomentum}, parse(Int, name[3:end]))
    elseif startswith(name, "po")
        return (ParticleStateful{Outgoing, Positron, SFourMomentum}, parse(Int, name[3:end]))
    else
        throw("Invalid name for a particle in the QED model")
    end
end

"""
    issame(T1::Type{<:ParticleStateful}, T2::Type{<:ParticleStateful})

For two given `ParticleStateful` types, return whether they are equivalent for the purpose of a Feynman Diagram. That means e.g. an `Incoming` `AntiFermion` is the same as an `Outgoing` `Fermion`. This is equivalent to `!caninteract(T1, T2)`.

See also: [`caninteract`](@ref) and [`interaction_result`](@ref)
"""
function issame(T1::Type{<:ParticleStateful}, T2::Type{<:ParticleStateful})
    return !caninteract(T1, T2)
end

"""
    QED_vertex()

Return the factor of a vertex in a QED feynman diagram.
"""
@inline function QED_vertex()::SLorentzVector{DiracMatrix}
    # Peskin-Schroeder notation
    return -1im * e * gamma()
end

@inline function QED_inner_edge(p::ParticleStateful)
    return propagator(particle_species(p), momentum(p))
end

"""
    QED_conserve_momentum(p1::ParticleStateful, p2::ParticleStateful)

Calculate and return a new particle from two given interacting ones at a vertex.
"""
function QED_conserve_momentum(
    p1::P1,
    p2::P2,
) where {
    Dir1 <: ParticleDirection,
    Dir2 <: ParticleDirection,
    Species1 <: AbstractParticleType,
    Species2 <: AbstractParticleType,
    P1 <: ParticleStateful{Dir1, Species1},
    P2 <: ParticleStateful{Dir2, Species2},
}
    P3 = interaction_result(P1, P2)
    p1_mom = momentum(p1)
    if (Dir1 <: Outgoing)
        p1_mom *= -1
    end
    p2_mom = momentum(p2)
    if (Dir2 <: Outgoing)
        p2_mom *= -1
    end

    p3_mom = p1_mom + p2_mom
    if (particle_direction(P3) isa Incoming)
        return ParticleStateful(particle_direction(P3), particle_species(P3), -p3_mom)
    end
    return ParticleStateful(particle_direction(P3), particle_species(P3), p3_mom)
end

"""
    model(::AbstractProcessDescription)

Return the model of this process description.
"""
model(::GenericQEDProcess) = QEDModel()
model(::PhaseSpacePoint) = QEDModel()

function get_particle(
    input::PhaseSpacePoint,
    t::Type{ParticleStateful{DIR, SPECIES}},
    n::Int,
) where {DIR <: ParticleDirection, SPECIES <: AbstractParticleType}
    i = 0
    for p in particles(input, DIR())
        if p isa t
            i += 1
            if i == n
                return p
            end
        end
    end
    @assert false "Invalid type given"
end
